EXPORTS
	com_right
        com_right_r
	free_error_table
	initialize_error_table_r
	add_to_error_table
	com_err
	com_err_va
	error_message
	error_table_name
	init_error_table
	reset_com_err_hook
	set_com_err_hook
        _et_list            DATA
