EXPORTS
	APOptions2int
	DigestTypes2int
	DistributionPointReasonFlags2int
	FastOptions2int
	KDCOptions2int
	KeyUsage2int
	SAMFlags2int
	TicketFlags2int
	_der_timegm
	_der_gmtime
	_heim_der_set_sort
	_heim_fix_dce
	_heim_len_int
	_heim_len_unsigned
	_heim_time2generalizedtime
	add_AuthorizationData
	add_CRLDistributionPoints
	add_DigestAlgorithmIdentifiers
	add_ETYPE_INFO
	add_ETYPE_INFO2
	add_Extensions
	add_GeneralNames
	add_METHOD_DATA
	add_Principals
	asn1_APOptions_units
	asn1_DigestTypes_units
	asn1_DistributionPointReasonFlags_units
	asn1_FastOptions_units
	asn1_KDCOptions_units
	asn1_KeyUsage_units
	asn1_SAMFlags_units
	asn1_TicketFlags_units
	asn1_oid_id_Userid	DATA
	asn1_oid_id_aes_128_cbc	DATA
	asn1_oid_id_aes_192_cbc	DATA
	asn1_oid_id_aes_256_cbc	DATA
	asn1_oid_id_at_commonName	DATA
	asn1_oid_id_at_countryName	DATA
	asn1_oid_id_at_generationQualifier	DATA
	asn1_oid_id_at_givenName	DATA
	asn1_oid_id_at_initials	DATA
	asn1_oid_id_at_localityName	DATA
	asn1_oid_id_at_name	DATA
	asn1_oid_id_at_organizationName	DATA
	asn1_oid_id_at_organizationalUnitName	DATA
	asn1_oid_id_at_pseudonym	DATA
	asn1_oid_id_at_serialNumber	DATA
	asn1_oid_id_at_stateOrProvinceName	DATA
	asn1_oid_id_at_streetAddress	DATA
	asn1_oid_id_at_surname	DATA
	asn1_oid_id_dhpublicnumber	DATA
	asn1_oid_id_domainComponent	DATA
	asn1_oid_id_dsa	DATA
	asn1_oid_id_dsa_with_sha1	DATA
	asn1_oid_id_ecDH	DATA
	asn1_oid_id_ecMQV	DATA
	asn1_oid_id_ecPublicKey	DATA
	asn1_oid_id_ec_group_secp160r1	DATA
	asn1_oid_id_ec_group_secp160r2	DATA
	asn1_oid_id_ec_group_secp224r1	DATA
	asn1_oid_id_ec_group_secp256r1	DATA
	asn1_oid_id_ec_group_secp384r1	DATA
	asn1_oid_id_ec_group_secp521r1	DATA
	asn1_oid_id_ecdsa_with_SHA1	DATA
	asn1_oid_id_ecdsa_with_SHA224   DATA
	asn1_oid_id_ecdsa_with_SHA256	DATA
	asn1_oid_id_ecdsa_with_SHA384   DATA
	asn1_oid_id_ecdsa_with_SHA512   DATA
	asn1_oid_id_heim_rsa_pkcs1_x509	DATA
	asn1_oid_id_ms_cert_enroll_domaincontroller	DATA
	asn1_oid_id_ms_client_authentication	DATA
	asn1_oid_id_netscape	DATA
	asn1_oid_id_netscape_cert_comment	DATA
	asn1_oid_id_nistAlgorithm	DATA
	asn1_oid_id_nist_aes_algs	DATA
	asn1_oid_id_nist_sha_algs	DATA
	asn1_oid_id_pbeWithSHAAnd128BitRC2_CBC	DATA
	asn1_oid_id_pbeWithSHAAnd128BitRC4	DATA
	asn1_oid_id_pbeWithSHAAnd2_KeyTripleDES_CBC	DATA
	asn1_oid_id_pbeWithSHAAnd3_KeyTripleDES_CBC	DATA
	asn1_oid_id_pbeWithSHAAnd40BitRC4	DATA
	asn1_oid_id_pbewithSHAAnd40BitRC2_CBC	DATA
	asn1_oid_id_pkauthdata	DATA
	asn1_oid_id_pkcs12_bagtypes	DATA
	asn1_oid_id_pkcs12_certBag	DATA
	asn1_oid_id_pkcs12_crlBag	DATA
	asn1_oid_id_pkcs12_keyBag	DATA
	asn1_oid_id_pkcs12_pkcs8ShroudedKeyBag	DATA
	asn1_oid_id_pkcs12_safeContentsBag	DATA
	asn1_oid_id_pkcs12_secretBag	DATA
	asn1_oid_id_pkcs1_md2WithRSAEncryption	DATA
	asn1_oid_id_pkcs1_md5WithRSAEncryption	DATA
	asn1_oid_id_pkcs1_rsaEncryption	DATA
	asn1_oid_id_pkcs1_sha1WithRSAEncryption	DATA
	asn1_oid_id_pkcs1_sha256WithRSAEncryption	DATA
	asn1_oid_id_pkcs1_sha384WithRSAEncryption	DATA
	asn1_oid_id_pkcs1_sha512WithRSAEncryption	DATA
	asn1_oid_id_pkcs2_md2	DATA
	asn1_oid_id_pkcs2_md4	DATA
	asn1_oid_id_pkcs2_md5	DATA
	asn1_oid_id_pkcs3_des_ede3_cbc	DATA
	asn1_oid_id_pkcs3_rc2_cbc	DATA
	asn1_oid_id_pkcs3_rc4	DATA
	asn1_oid_id_pkcs7	DATA
	asn1_oid_id_pkcs7_data	DATA
	asn1_oid_id_pkcs7_digestedData	DATA
	asn1_oid_id_pkcs7_encryptedData	DATA
	asn1_oid_id_pkcs7_envelopedData	DATA
	asn1_oid_id_pkcs7_signedAndEnvelopedData	DATA
	asn1_oid_id_pkcs7_signedData	DATA
	asn1_oid_id_pkcs9_contentType	DATA
	asn1_oid_id_pkcs9_countersignature	DATA
	asn1_oid_id_pkcs9_emailAddress	DATA
	asn1_oid_id_pkcs9_messageDigest	DATA
	asn1_oid_id_pkcs9_signingTime	DATA
	asn1_oid_id_pkcs_1	DATA
	asn1_oid_id_pkcs_12	DATA
	asn1_oid_id_pkcs_12PbeIds	DATA
	asn1_oid_id_pkcs_2	DATA
	asn1_oid_id_pkcs_3	DATA
	asn1_oid_id_pkcs_9	DATA
	asn1_oid_id_pkcs_9_at_certTypes	DATA
	asn1_oid_id_pkcs_9_at_certTypes_x509	DATA
	asn1_oid_id_pkcs_9_at_friendlyName	DATA
	asn1_oid_id_pkcs_9_at_localKeyId	DATA
	asn1_oid_id_pkdhkeydata	DATA
	asn1_oid_id_pkekuoid	DATA
	asn1_oid_id_pkinit	DATA
	asn1_oid_id_pkinit_kdf	DATA
	asn1_oid_id_pkinit_kdf_ah_sha1	DATA
	asn1_oid_id_pkinit_kdf_ah_sha256	DATA
	asn1_oid_id_pkinit_kdf_ah_sha512	DATA
	asn1_oid_id_pkinit_ms_eku	DATA
	asn1_oid_id_pkinit_ms_san	DATA
	asn1_oid_id_pkinit_san	DATA
	asn1_oid_id_pkix	DATA
	asn1_oid_id_pkix_kp	DATA
	asn1_oid_id_pkix_kp_OCSPSigning	DATA
	asn1_oid_id_pkix_kp_clientAuth	DATA
	asn1_oid_id_pkix_kp_emailProtection	DATA
	asn1_oid_id_pkix_kp_serverAuth	DATA
	asn1_oid_id_pkix_kp_timeStamping	DATA
	asn1_oid_id_pkix_on	DATA
	asn1_oid_id_pkix_on_dnsSRV	DATA
	asn1_oid_id_pkix_on_xmppAddr	DATA
	asn1_oid_id_pkix_pe	DATA
	asn1_oid_id_pkix_pe_authorityInfoAccess	DATA
	asn1_oid_id_pkix_pe_proxyCertInfo	DATA
	asn1_oid_id_pkix_ppl	DATA
	asn1_oid_id_pkix_ppl_anyLanguage	DATA
	asn1_oid_id_pkix_ppl_independent	DATA
	asn1_oid_id_pkix_ppl_inheritAll	DATA
	asn1_oid_id_pkkdcekuoid	DATA
	asn1_oid_id_pkrkeydata	DATA
	asn1_oid_id_rsa_digestAlgorithm	DATA
	asn1_oid_id_rsa_digest_md2	DATA
	asn1_oid_id_rsa_digest_md4	DATA
	asn1_oid_id_rsa_digest_md5	DATA
	asn1_oid_id_rsadsi_des_ede3_cbc	DATA
	asn1_oid_id_rsadsi_encalg	DATA
	asn1_oid_id_rsadsi_rc2_cbc	DATA
	asn1_oid_id_secsig_sha_1	DATA
	asn1_oid_id_secsig_sha_1WithRSAEncryption	DATA
	asn1_oid_id_sha224	DATA
	asn1_oid_id_sha256	DATA
	asn1_oid_id_sha384	DATA
	asn1_oid_id_sha512	DATA
	asn1_oid_id_uspkicommon_card_id	DATA
	asn1_oid_id_uspkicommon_piv_interim	DATA
	asn1_oid_id_x509_ce	DATA
	asn1_oid_id_x509_ce_authorityKeyIdentifier	DATA
	asn1_oid_id_x509_ce_basicConstraints	DATA
	asn1_oid_id_x509_ce_cRLDistributionPoints	DATA
	asn1_oid_id_x509_ce_cRLNumber	DATA
	asn1_oid_id_x509_ce_cRLReason	DATA
	asn1_oid_id_x509_ce_certificateIssuer	DATA
	asn1_oid_id_x509_ce_certificatePolicies	DATA
	asn1_oid_id_x509_ce_deltaCRLIndicator	DATA
	asn1_oid_id_x509_ce_extKeyUsage	DATA
	asn1_oid_id_x509_ce_freshestCRL	DATA
	asn1_oid_id_x509_ce_holdInstructionCode	DATA
	asn1_oid_id_x509_ce_inhibitAnyPolicy	DATA
	asn1_oid_id_x509_ce_invalidityDate	DATA
	asn1_oid_id_x509_ce_issuerAltName	DATA
	asn1_oid_id_x509_ce_issuingDistributionPoint	DATA
	asn1_oid_id_x509_ce_keyUsage	DATA
	asn1_oid_id_x509_ce_nameConstraints	DATA
	asn1_oid_id_x509_ce_policyConstraints	DATA
	asn1_oid_id_x509_ce_policyMappings	DATA
	asn1_oid_id_x509_ce_privateKeyUsagePeriod	DATA
	asn1_oid_id_x509_ce_subjectAltName	DATA
	asn1_oid_id_x509_ce_subjectDirectoryAttributes	DATA
	asn1_oid_id_x509_ce_subjectKeyIdentifier	DATA
	asn1_oid_id_x520_at	DATA
	asn1_oid_id_x9_57	DATA
	copy_AD_AND_OR
	copy_AD_IF_RELEVANT
	copy_AD_INITIAL_VERIFIED_CAS
	copy_AD_KDCIssued
	copy_AD_LoginAlias
	copy_AD_MANDATORY_FOR_KDC
	copy_APOptions
	copy_AP_REP
	copy_AP_REQ
	copy_AS_REP
	copy_AS_REQ
	copy_AUTHDATA_TYPE
	copy_AccessDescription
	copy_AlgorithmIdentifier
	copy_Attribute
	copy_AttributeType
	copy_AttributeTypeAndValue
	copy_AttributeValue
	copy_AuthPack
	copy_AuthPack_Win2k
	copy_Authenticator
	copy_AuthorityInfoAccessSyntax
	copy_AuthorityKeyIdentifier
	copy_AuthorizationData
	copy_AuthorizationDataElement
	copy_BaseDistance
	copy_BasicConstraints
	copy_CKSUMTYPE
	copy_CMSAttributes
	copy_CMSCBCParameter
	copy_CMSEncryptedData
	copy_CMSIdentifier
	copy_CMSRC2CBCParameter
	copy_CMSVersion
	copy_CRLCertificateList
	copy_CRLDistributionPoints
	copy_CRLReason
	copy_Certificate
	copy_CertificateList
	copy_CertificateRevocationLists
	copy_CertificateSerialNumber
	copy_CertificateSet
	copy_Certificates
	copy_ChangePasswdDataMS
	copy_Checksum
	copy_ContentEncryptionAlgorithmIdentifier
	copy_ContentInfo
	copy_ContentType
	copy_DHNonce
	copy_DHParameter
	copy_DHPublicKey
	copy_DHRepInfo
	copy_DSAParams
	copy_DSAPublicKey
	copy_DSASigValue
	copy_DigestAlgorithmIdentifier
	copy_DigestAlgorithmIdentifiers
	copy_DigestError
	copy_DigestInfo
	copy_DigestInit
	copy_DigestInitReply
	copy_DigestREP
	copy_DigestREQ
	copy_DigestRepInner
	copy_DigestReqInner
	copy_DigestRequest
	copy_DigestResponse
	copy_DigestTypes
	copy_DirectoryString
	copy_DistributionPoint
	copy_DistributionPointName
	copy_DistributionPointReasonFlags
	copy_DomainParameters
	copy_ECDSA_Sig_Value
	copy_ECParameters
	copy_ECPoint
	copy_ENCTYPE
	copy_ETYPE_INFO
	copy_ETYPE_INFO2
	copy_ETYPE_INFO2_ENTRY
	copy_ETYPE_INFO_ENTRY
	copy_EncAPRepPart
	copy_EncASRepPart
	copy_EncKDCRepPart
	copy_EncKrbCredPart
	copy_EncKrbPrivPart
	copy_EncTGSRepPart
	copy_EncTicketPart
	copy_EncapsulatedContentInfo
	copy_EncryptedContent
	copy_EncryptedContentInfo
	copy_EncryptedData
	copy_EncryptedKey
	copy_EncryptionKey
	copy_EnvelopedData
	copy_EtypeList
	copy_ExtKeyUsage
	copy_Extension
	copy_Extensions
	copy_ExternalPrincipalIdentifier
	copy_ExternalPrincipalIdentifiers
	copy_FastOptions
	copy_GeneralName
	copy_GeneralNames
	copy_GeneralSubtree
	copy_GeneralSubtrees
	copy_HostAddress
	copy_HostAddresses
	copy_IssuerAndSerialNumber
	copy_KDCDHKeyInfo
	copy_KDCDHKeyInfo_Win2k
	copy_KDCFastCookie
	copy_KDCFastState
	copy_KDCOptions
	copy_KDC_REP
	copy_KDC_REQ
	copy_KDC_REQ_BODY
	copy_KDFAlgorithmId
	copy_KRB5PrincipalName
	copy_KRB5SignedPath
	copy_KRB5SignedPathData
	copy_KRB_CRED
	copy_KRB_ERROR
	copy_KRB_PRIV
	copy_KRB_SAFE
	copy_KRB_SAFE_BODY
	copy_KX509_ERROR_CODE
	copy_KerberosString
	copy_KerberosTime
	copy_KeyEncryptionAlgorithmIdentifier
	copy_KeyIdentifier
	copy_KeyTransRecipientInfo
	copy_KeyUsage
	copy_KrbCredInfo
	copy_KrbFastArmor
	copy_KrbFastArmoredRep
	copy_KrbFastArmoredReq
	copy_KrbFastFinished
	copy_KrbFastReq
	copy_KrbFastResponse
	copy_Kx509Request
	copy_Kx509Response
	copy_LR_TYPE
	copy_LastReq
	copy_MESSAGE_TYPE
	copy_METHOD_DATA
	copy_MS_UPN_SAN
	copy_MessageDigest
	copy_NAME_TYPE
	copy_NTLMInit
	copy_NTLMInitReply
	copy_NTLMReply
	copy_NTLMRequest
	copy_NTLMRequest2
	copy_NTLMResponse
	copy_Name
	copy_NameConstraints
	copy_OriginatorInfo
	copy_OtherName
	copy_PADATA_TYPE
	copy_PA_DATA
	copy_PA_ENC_SAM_RESPONSE_ENC
	copy_PA_ENC_TS_ENC
	copy_PA_FX_FAST_REPLY
	copy_PA_FX_FAST_REQUEST
	copy_PA_PAC_REQUEST
	copy_PA_PK_AS_REP
	copy_PA_PK_AS_REP_BTMM
	copy_PA_PK_AS_REP_Win2k
	copy_PA_PK_AS_REQ
	copy_PA_PK_AS_REQ_Win2k
	copy_PA_S4U2Self
	copy_PA_SAM_CHALLENGE_2
	copy_PA_SAM_CHALLENGE_2_BODY
	copy_PA_SAM_REDIRECT
	copy_PA_SAM_RESPONSE_2
	copy_PA_SAM_TYPE
	copy_PA_SERVER_REFERRAL_DATA
	copy_PA_ServerReferralData
	copy_PA_SvrReferralData
	copy_PKAuthenticator
	copy_PKAuthenticator_Win2k
	copy_PKCS12_Attribute
	copy_PKCS12_Attributes
	copy_PKCS12_AuthenticatedSafe
	copy_PKCS12_CertBag
	copy_PKCS12_MacData
	copy_PKCS12_OctetString
	copy_PKCS12_PBEParams
	copy_PKCS12_PFX
	copy_PKCS12_SafeBag
	copy_PKCS12_SafeContents
	copy_PKCS8Attributes
	copy_PKCS8EncryptedData
	copy_PKCS8EncryptedPrivateKeyInfo
	copy_PKCS8PrivateKey
	copy_PKCS8PrivateKeyAlgorithmIdentifier
	copy_PKCS8PrivateKeyInfo
	copy_PKCS9_BMPString
	copy_PKCS9_friendlyName
	copy_PKIXXmppAddr
	copy_PROV_SRV_LOCATION
	copy_PkinitSP80056AOtherInfo
	copy_PkinitSuppPubInfo
	copy_Principal
	copy_PrincipalName
	copy_Principals
	copy_ProxyCertInfo
	copy_ProxyPolicy
	copy_RDNSequence
	copy_RSAPrivateKey
	copy_RSAPublicKey
	copy_Realm
	copy_RecipientIdentifier
	copy_RecipientInfo
	copy_RecipientInfos
	copy_RelativeDistinguishedName
	copy_ReplyKeyPack
	copy_ReplyKeyPack_Win2k
	copy_SAMFlags
	copy_SignatureAlgorithmIdentifier
	copy_SignatureValue
	copy_SignedData
	copy_SignerIdentifier
	copy_SignerInfo
	copy_SignerInfos
	copy_SubjectKeyIdentifier
	copy_SubjectPublicKeyInfo
	copy_TBSCRLCertList
	copy_TBSCertificate
	copy_TD_DH_PARAMETERS
	copy_TD_INVALID_CERTIFICATES
	copy_TD_TRUSTED_CERTIFIERS
	copy_TGS_REP
	copy_TGS_REQ
	copy_TYPED_DATA
	copy_Ticket
	copy_TicketFlags
	copy_Time
	copy_TransitedEncoding
	copy_TrustedCA
	copy_TrustedCA_Win2k
	copy_TypedData
	copy_UniqueIdentifier
	copy_UnprotectedAttributes
	copy_ValidationParms
	copy_Validity
	copy_Version
	copy_heim_any
	copy_heim_any_set
	copy_krb5int32
	copy_krb5uint32
	decode_AD_AND_OR
	decode_AD_IF_RELEVANT
	decode_AD_INITIAL_VERIFIED_CAS
	decode_AD_KDCIssued
	decode_AD_LoginAlias
	decode_AD_MANDATORY_FOR_KDC
	decode_APOptions
	decode_AP_REP
	decode_AP_REQ
	decode_AS_REP
	decode_AS_REQ
	decode_AUTHDATA_TYPE
	decode_AccessDescription
	decode_AlgorithmIdentifier
	decode_Attribute
	decode_AttributeType
	decode_AttributeTypeAndValue
	decode_AttributeValue
	decode_AuthPack
	decode_AuthPack_Win2k
	decode_Authenticator
	decode_AuthorityInfoAccessSyntax
	decode_AuthorityKeyIdentifier
	decode_AuthorizationData
	decode_AuthorizationDataElement
	decode_BaseDistance
	decode_BasicConstraints
	decode_CKSUMTYPE
	decode_CMSAttributes
	decode_CMSCBCParameter
	decode_CMSEncryptedData
	decode_CMSIdentifier
	decode_CMSRC2CBCParameter
	decode_CMSVersion
	decode_CRLCertificateList
	decode_CRLDistributionPoints
	decode_CRLReason
	decode_Certificate
	decode_CertificateList
	decode_CertificateRevocationLists
	decode_CertificateSerialNumber
	decode_CertificateSet
	decode_Certificates
	decode_ChangePasswdDataMS
	decode_Checksum
	decode_ContentEncryptionAlgorithmIdentifier
	decode_ContentInfo
	decode_ContentType
	decode_DHNonce
	decode_DHParameter
	decode_DHPublicKey
	decode_DHRepInfo
	decode_DSAParams
	decode_DSAPublicKey
	decode_DSASigValue
	decode_DigestAlgorithmIdentifier
	decode_DigestAlgorithmIdentifiers
	decode_DigestError
	decode_DigestInfo
	decode_DigestInit
	decode_DigestInitReply
	decode_DigestREP
	decode_DigestREQ
	decode_DigestRepInner
	decode_DigestReqInner
	decode_DigestRequest
	decode_DigestResponse
	decode_DigestTypes
	decode_DirectoryString
	decode_DistributionPoint
	decode_DistributionPointName
	decode_DistributionPointReasonFlags
	decode_DomainParameters
	decode_ECDSA_Sig_Value
	decode_ECParameters
	decode_ECPoint
	decode_ENCTYPE
	decode_ETYPE_INFO
	decode_ETYPE_INFO2
	decode_ETYPE_INFO2_ENTRY
	decode_ETYPE_INFO_ENTRY
	decode_EncAPRepPart
	decode_EncASRepPart
	decode_EncKDCRepPart
	decode_EncKrbCredPart
	decode_EncKrbPrivPart
	decode_EncTGSRepPart
	decode_EncTicketPart
	decode_EncapsulatedContentInfo
	decode_EncryptedContent
	decode_EncryptedContentInfo
	decode_EncryptedData
	decode_EncryptedKey
	decode_EncryptionKey
	decode_EnvelopedData
	decode_EtypeList
	decode_ExtKeyUsage
	decode_Extension
	decode_Extensions
	decode_ExternalPrincipalIdentifier
	decode_ExternalPrincipalIdentifiers
	decode_FastOptions
	decode_GeneralName
	decode_GeneralNames
	decode_GeneralSubtree
	decode_GeneralSubtrees
	decode_HostAddress
	decode_HostAddresses
	decode_IssuerAndSerialNumber
	decode_KDCDHKeyInfo
	decode_KDCDHKeyInfo_Win2k
	decode_KDCFastCookie
	decode_KDCFastState
	decode_KDCOptions
	decode_KDC_REP
	decode_KDC_REQ
	decode_KDC_REQ_BODY
	decode_KDFAlgorithmId
	decode_KRB5PrincipalName
	decode_KRB5SignedPath
	decode_KRB5SignedPathData
	decode_KRB_CRED
	decode_KRB_ERROR
	decode_KRB_PRIV
	decode_KRB_SAFE
	decode_KRB_SAFE_BODY
	decode_KX509_ERROR_CODE
	decode_KerberosString
	decode_KerberosTime
	decode_KeyEncryptionAlgorithmIdentifier
	decode_KeyIdentifier
	decode_KeyTransRecipientInfo
	decode_KeyUsage
	decode_KrbCredInfo
	decode_KrbFastArmor
	decode_KrbFastArmoredRep
	decode_KrbFastArmoredReq
	decode_KrbFastFinished
	decode_KrbFastReq
	decode_KrbFastResponse
	decode_Kx509Request
	decode_Kx509Response
	decode_LR_TYPE
	decode_LastReq
	decode_MESSAGE_TYPE
	decode_METHOD_DATA
	decode_MS_UPN_SAN
	decode_MessageDigest
	decode_NAME_TYPE
	decode_NTLMInit
	decode_NTLMInitReply
	decode_NTLMReply
	decode_NTLMRequest
	decode_NTLMRequest2
	decode_NTLMResponse
	decode_Name
	decode_NameConstraints
	decode_OriginatorInfo
	decode_OtherName
	decode_PADATA_TYPE
	decode_PA_DATA
	decode_PA_ENC_SAM_RESPONSE_ENC
	decode_PA_ENC_TS_ENC
	decode_PA_FX_FAST_REPLY
	decode_PA_FX_FAST_REQUEST
	decode_PA_PAC_REQUEST
	decode_PA_PK_AS_REP
	decode_PA_PK_AS_REP_BTMM
	decode_PA_PK_AS_REP_Win2k
	decode_PA_PK_AS_REQ
	decode_PA_PK_AS_REQ_Win2k
	decode_PA_S4U2Self
	decode_PA_SAM_CHALLENGE_2
	decode_PA_SAM_CHALLENGE_2_BODY
	decode_PA_SAM_REDIRECT
	decode_PA_SAM_RESPONSE_2
	decode_PA_SAM_TYPE
	decode_PA_SERVER_REFERRAL_DATA
	decode_PA_ServerReferralData
	decode_PA_SvrReferralData
	decode_PKAuthenticator
	decode_PKAuthenticator_Win2k
	decode_PKCS12_Attribute
	decode_PKCS12_Attributes
	decode_PKCS12_AuthenticatedSafe
	decode_PKCS12_CertBag
	decode_PKCS12_MacData
	decode_PKCS12_OctetString
	decode_PKCS12_PBEParams
	decode_PKCS12_PFX
	decode_PKCS12_SafeBag
	decode_PKCS12_SafeContents
	decode_PKCS8Attributes
	decode_PKCS8EncryptedData
	decode_PKCS8EncryptedPrivateKeyInfo
	decode_PKCS8PrivateKey
	decode_PKCS8PrivateKeyAlgorithmIdentifier
	decode_PKCS8PrivateKeyInfo
	decode_PKCS9_BMPString
	decode_PKCS9_friendlyName
	decode_PKIXXmppAddr
	decode_PROV_SRV_LOCATION
	decode_PkinitSP80056AOtherInfo
	decode_PkinitSuppPubInfo
	decode_Principal
	decode_PrincipalName
	decode_Principals
	decode_ProxyCertInfo
	decode_ProxyPolicy
	decode_RDNSequence
	decode_RSAPrivateKey
	decode_RSAPublicKey
	decode_Realm
	decode_RecipientIdentifier
	decode_RecipientInfo
	decode_RecipientInfos
	decode_RelativeDistinguishedName
	decode_ReplyKeyPack
	decode_ReplyKeyPack_Win2k
	decode_SAMFlags
	decode_SignatureAlgorithmIdentifier
	decode_SignatureValue
	decode_SignedData
	decode_SignerIdentifier
	decode_SignerInfo
	decode_SignerInfos
	decode_SubjectKeyIdentifier
	decode_SubjectPublicKeyInfo
	decode_TBSCRLCertList
	decode_TBSCertificate
	decode_TD_DH_PARAMETERS
	decode_TD_INVALID_CERTIFICATES
	decode_TD_TRUSTED_CERTIFIERS
	decode_TGS_REP
	decode_TGS_REQ
	decode_TYPED_DATA
	decode_Ticket
	decode_TicketFlags
	decode_Time
	decode_TransitedEncoding
	decode_TrustedCA
	decode_TrustedCA_Win2k
	decode_TypedData
	decode_UniqueIdentifier
	decode_UnprotectedAttributes
	decode_ValidationParms
	decode_Validity
	decode_Version
	decode_heim_any
	decode_heim_any_set
	decode_krb5int32
	decode_krb5uint32
	der_copy_bit_string
	der_copy_bmp_string
	der_copy_general_string
	der_copy_generalized_time
	der_copy_heim_integer
	der_copy_ia5_string
	der_copy_integer
	der_copy_integer64
	der_copy_octet_string
	der_copy_oid
	der_copy_printable_string
	der_copy_universal_string
	der_copy_unsigned
	der_copy_unsigned64
	der_copy_utctime
	der_copy_utf8string
	der_copy_visible_string
	der_free_bit_string
	der_free_bmp_string
	der_free_general_string
	der_free_generalized_time
	der_free_heim_integer
	der_free_ia5_string
	der_free_integer
	der_free_integer64
	der_free_octet_string
	der_free_oid
	der_free_printable_string
	der_free_universal_string
	der_free_unsigned
	der_free_unsigned64
	der_free_utctime
	der_free_utf8string
	der_free_visible_string
	der_get_bit_string
	der_get_bmp_string
	der_get_boolean
	der_get_class_name
	der_get_class_num
	der_get_general_string
	der_get_generalized_time
	der_get_heim_integer
	der_get_ia5_string
	der_get_integer
	der_get_integer64
	der_get_length
	der_get_octet_string
	der_get_octet_string_ber
	der_get_oid
	der_get_printable_string
	der_get_tag
	der_get_tag_name
	der_get_tag_num
	der_get_type_name
	der_get_type_num
	der_get_universal_string
	der_get_unsigned
	der_get_unsigned64
	der_get_utctime
	der_get_utf8string
	der_get_visible_string
	der_heim_bit_string_cmp
	der_heim_bmp_string_cmp
	der_heim_integer_cmp
	der_heim_octet_string_cmp
	der_heim_oid_cmp
	der_heim_universal_string_cmp
	der_ia5_string_cmp
	der_length_bit_string
	der_length_bmp_string
	der_length_boolean
	der_length_enumerated
	der_length_general_string
	der_length_generalized_time
	der_length_heim_integer
	der_length_ia5_string
	der_length_integer
	der_length_integer64
	der_length_len
	der_length_octet_string
	der_length_oid
	der_length_printable_string
	der_length_tag
	der_length_universal_string
	der_length_unsigned
	der_length_unsigned64
	der_length_utctime
	der_length_utf8string
	der_length_visible_string
	der_match_tag
	der_match_tag2
	der_match_tag_and_length
	der_parse_heim_oid
	der_parse_hex_heim_integer
	der_print_heim_oid
	der_print_hex_heim_integer
	der_printable_string_cmp
	der_put_bit_string
	der_put_bmp_string
	der_put_boolean
	der_put_general_string
	der_put_generalized_time
	der_put_heim_integer
	der_put_ia5_string
	der_put_integer
	der_put_integer64
	der_put_length
	der_put_length_and_tag
	der_put_octet_string
	der_put_oid
	der_put_printable_string
	der_put_tag
	der_put_universal_string
	der_put_unsigned
	der_put_unsigned64
	der_put_utctime
	der_put_utf8string
	der_put_visible_string
	encode_AD_AND_OR
	encode_AD_IF_RELEVANT
	encode_AD_INITIAL_VERIFIED_CAS
	encode_AD_KDCIssued
	encode_AD_LoginAlias
	encode_AD_MANDATORY_FOR_KDC
	encode_APOptions
	encode_AP_REP
	encode_AP_REQ
	encode_AS_REP
	encode_AS_REQ
	encode_AUTHDATA_TYPE
	encode_AccessDescription
	encode_AlgorithmIdentifier
	encode_Attribute
	encode_AttributeType
	encode_AttributeTypeAndValue
	encode_AttributeValue
	encode_AuthPack
	encode_AuthPack_Win2k
	encode_Authenticator
	encode_AuthorityInfoAccessSyntax
	encode_AuthorityKeyIdentifier
	encode_AuthorizationData
	encode_AuthorizationDataElement
	encode_BaseDistance
	encode_BasicConstraints
	encode_CKSUMTYPE
	encode_CMSAttributes
	encode_CMSCBCParameter
	encode_CMSEncryptedData
	encode_CMSIdentifier
	encode_CMSRC2CBCParameter
	encode_CMSVersion
	encode_CRLCertificateList
	encode_CRLDistributionPoints
	encode_CRLReason
	encode_Certificate
	encode_CertificateList
	encode_CertificateRevocationLists
	encode_CertificateSerialNumber
	encode_CertificateSet
	encode_Certificates
	encode_ChangePasswdDataMS
	encode_Checksum
	encode_ContentEncryptionAlgorithmIdentifier
	encode_ContentInfo
	encode_ContentType
	encode_DHNonce
	encode_DHParameter
	encode_DHPublicKey
	encode_DHRepInfo
	encode_DSAParams
	encode_DSAPublicKey
	encode_DSASigValue
	encode_DigestAlgorithmIdentifier
	encode_DigestAlgorithmIdentifiers
	encode_DigestError
	encode_DigestInfo
	encode_DigestInit
	encode_DigestInitReply
	encode_DigestREP
	encode_DigestREQ
	encode_DigestRepInner
	encode_DigestReqInner
	encode_DigestRequest
	encode_DigestResponse
	encode_DigestTypes
	encode_DirectoryString
	encode_DistributionPoint
	encode_DistributionPointName
	encode_DistributionPointReasonFlags
	encode_DomainParameters
	encode_ECDSA_Sig_Value
	encode_ECParameters
	encode_ECPoint
	encode_ENCTYPE
	encode_ETYPE_INFO
	encode_ETYPE_INFO2
	encode_ETYPE_INFO2_ENTRY
	encode_ETYPE_INFO_ENTRY
	encode_EncAPRepPart
	encode_EncASRepPart
	encode_EncKDCRepPart
	encode_EncKrbCredPart
	encode_EncKrbPrivPart
	encode_EncTGSRepPart
	encode_EncTicketPart
	encode_EncapsulatedContentInfo
	encode_EncryptedContent
	encode_EncryptedContentInfo
	encode_EncryptedData
	encode_EncryptedKey
	encode_EncryptionKey
	encode_EnvelopedData
	encode_EtypeList
	encode_ExtKeyUsage
	encode_Extension
	encode_Extensions
	encode_ExternalPrincipalIdentifier
	encode_ExternalPrincipalIdentifiers
	encode_FastOptions
	encode_GeneralName
	encode_GeneralNames
	encode_GeneralSubtree
	encode_GeneralSubtrees
	encode_HostAddress
	encode_HostAddresses
	encode_IssuerAndSerialNumber
	encode_KDCDHKeyInfo
	encode_KDCDHKeyInfo_Win2k
	encode_KDCFastCookie
	encode_KDCFastState
	encode_KDCOptions
	encode_KDC_REP
	encode_KDC_REQ
	encode_KDC_REQ_BODY
	encode_KDFAlgorithmId
	encode_KRB5PrincipalName
	encode_KRB5SignedPath
	encode_KRB5SignedPathData
	encode_KRB_CRED
	encode_KRB_ERROR
	encode_KRB_PRIV
	encode_KRB_SAFE
	encode_KRB_SAFE_BODY
	encode_KX509_ERROR_CODE
	encode_KerberosString
	encode_KerberosTime
	encode_KeyEncryptionAlgorithmIdentifier
	encode_KeyIdentifier
	encode_KeyTransRecipientInfo
	encode_KeyUsage
	encode_KrbCredInfo
	encode_KrbFastArmor
	encode_KrbFastArmoredRep
	encode_KrbFastArmoredReq
	encode_KrbFastFinished
	encode_KrbFastReq
	encode_KrbFastResponse
	encode_Kx509Request
	encode_Kx509Response
	encode_LR_TYPE
	encode_LastReq
	encode_MESSAGE_TYPE
	encode_METHOD_DATA
	encode_MS_UPN_SAN
	encode_MessageDigest
	encode_NAME_TYPE
	encode_NTLMInit
	encode_NTLMInitReply
	encode_NTLMReply
	encode_NTLMRequest
	encode_NTLMRequest2
	encode_NTLMResponse
	encode_Name
	encode_NameConstraints
	encode_OriginatorInfo
	encode_OtherName
	encode_PADATA_TYPE
	encode_PA_DATA
	encode_PA_ENC_SAM_RESPONSE_ENC
	encode_PA_ENC_TS_ENC
	encode_PA_FX_FAST_REPLY
	encode_PA_FX_FAST_REQUEST
	encode_PA_PAC_REQUEST
	encode_PA_PK_AS_REP
	encode_PA_PK_AS_REP_BTMM
	encode_PA_PK_AS_REP_Win2k
	encode_PA_PK_AS_REQ
	encode_PA_PK_AS_REQ_Win2k
	encode_PA_S4U2Self
	encode_PA_SAM_CHALLENGE_2
	encode_PA_SAM_CHALLENGE_2_BODY
	encode_PA_SAM_REDIRECT
	encode_PA_SAM_RESPONSE_2
	encode_PA_SAM_TYPE
	encode_PA_SERVER_REFERRAL_DATA
	encode_PA_ServerReferralData
	encode_PA_SvrReferralData
	encode_PKAuthenticator
	encode_PKAuthenticator_Win2k
	encode_PKCS12_Attribute
	encode_PKCS12_Attributes
	encode_PKCS12_AuthenticatedSafe
	encode_PKCS12_CertBag
	encode_PKCS12_MacData
	encode_PKCS12_OctetString
	encode_PKCS12_PBEParams
	encode_PKCS12_PFX
	encode_PKCS12_SafeBag
	encode_PKCS12_SafeContents
	encode_PKCS8Attributes
	encode_PKCS8EncryptedData
	encode_PKCS8EncryptedPrivateKeyInfo
	encode_PKCS8PrivateKey
	encode_PKCS8PrivateKeyAlgorithmIdentifier
	encode_PKCS8PrivateKeyInfo
	encode_PKCS9_BMPString
	encode_PKCS9_friendlyName
	encode_PKIXXmppAddr
	encode_PROV_SRV_LOCATION
	encode_PkinitSP80056AOtherInfo
	encode_PkinitSuppPubInfo
	encode_Principal
	encode_PrincipalName
	encode_Principals
	encode_ProxyCertInfo
	encode_ProxyPolicy
	encode_RDNSequence
	encode_RSAPrivateKey
	encode_RSAPublicKey
	encode_Realm
	encode_RecipientIdentifier
	encode_RecipientInfo
	encode_RecipientInfos
	encode_RelativeDistinguishedName
	encode_ReplyKeyPack
	encode_ReplyKeyPack_Win2k
	encode_SAMFlags
	encode_SignatureAlgorithmIdentifier
	encode_SignatureValue
	encode_SignedData
	encode_SignerIdentifier
	encode_SignerInfo
	encode_SignerInfos
	encode_SubjectKeyIdentifier
	encode_SubjectPublicKeyInfo
	encode_TBSCRLCertList
	encode_TBSCertificate
	encode_TD_DH_PARAMETERS
	encode_TD_INVALID_CERTIFICATES
	encode_TD_TRUSTED_CERTIFIERS
	encode_TGS_REP
	encode_TGS_REQ
	encode_TYPED_DATA
	encode_Ticket
	encode_TicketFlags
	encode_Time
	encode_TransitedEncoding
	encode_TrustedCA
	encode_TrustedCA_Win2k
	encode_TypedData
	encode_UniqueIdentifier
	encode_UnprotectedAttributes
	encode_ValidationParms
	encode_Validity
	encode_Version
	encode_heim_any
	encode_heim_any_set
	encode_krb5int32
	encode_krb5uint32
	free_AD_AND_OR
	free_AD_IF_RELEVANT
	free_AD_INITIAL_VERIFIED_CAS
	free_AD_KDCIssued
	free_AD_LoginAlias
	free_AD_MANDATORY_FOR_KDC
	free_APOptions
	free_AP_REP
	free_AP_REQ
	free_AS_REP
	free_AS_REQ
	free_AUTHDATA_TYPE
	free_AccessDescription
	free_AlgorithmIdentifier
	free_Attribute
	free_AttributeType
	free_AttributeTypeAndValue
	free_AttributeValue
	free_AuthPack
	free_AuthPack_Win2k
	free_Authenticator
	free_AuthorityInfoAccessSyntax
	free_AuthorityKeyIdentifier
	free_AuthorizationData
	free_AuthorizationDataElement
	free_BaseDistance
	free_BasicConstraints
	free_CKSUMTYPE
	free_CMSAttributes
	free_CMSCBCParameter
	free_CMSEncryptedData
	free_CMSIdentifier
	free_CMSRC2CBCParameter
	free_CMSVersion
	free_CRLCertificateList
	free_CRLDistributionPoints
	free_CRLReason
	free_Certificate
	free_CertificateList
	free_CertificateRevocationLists
	free_CertificateSerialNumber
	free_CertificateSet
	free_Certificates
	free_ChangePasswdDataMS
	free_Checksum
	free_ContentEncryptionAlgorithmIdentifier
	free_ContentInfo
	free_ContentType
	free_DHNonce
	free_DHParameter
	free_DHPublicKey
	free_DHRepInfo
	free_DSAParams
	free_DSAPublicKey
	free_DSASigValue
	free_DigestAlgorithmIdentifier
	free_DigestAlgorithmIdentifiers
	free_DigestError
	free_DigestInfo
	free_DigestInit
	free_DigestInitReply
	free_DigestREP
	free_DigestREQ
	free_DigestRepInner
	free_DigestReqInner
	free_DigestRequest
	free_DigestResponse
	free_DigestTypes
	free_DirectoryString
	free_DistributionPoint
	free_DistributionPointName
	free_DistributionPointReasonFlags
	free_DomainParameters
	free_ECDSA_Sig_Value
	free_ECParameters
	free_ECPoint
	free_ENCTYPE
	free_ETYPE_INFO
	free_ETYPE_INFO2
	free_ETYPE_INFO2_ENTRY
	free_ETYPE_INFO_ENTRY
	free_EncAPRepPart
	free_EncASRepPart
	free_EncKDCRepPart
	free_EncKrbCredPart
	free_EncKrbPrivPart
	free_EncTGSRepPart
	free_EncTicketPart
	free_EncapsulatedContentInfo
	free_EncryptedContent
	free_EncryptedContentInfo
	free_EncryptedData
	free_EncryptedKey
	free_EncryptionKey
	free_EnvelopedData
	free_EtypeList
	free_ExtKeyUsage
	free_Extension
	free_Extensions
	free_ExternalPrincipalIdentifier
	free_ExternalPrincipalIdentifiers
	free_FastOptions
	free_GeneralName
	free_GeneralNames
	free_GeneralSubtree
	free_GeneralSubtrees
	free_HostAddress
	free_HostAddresses
	free_IssuerAndSerialNumber
	free_KDCDHKeyInfo
	free_KDCDHKeyInfo_Win2k
	free_KDCFastCookie
	free_KDCFastState
	free_KDCOptions
	free_KDC_REP
	free_KDC_REQ
	free_KDC_REQ_BODY
	free_KDFAlgorithmId
	free_KRB5PrincipalName
	free_KRB5SignedPath
	free_KRB5SignedPathData
	free_KRB_CRED
	free_KRB_ERROR
	free_KRB_PRIV
	free_KRB_SAFE
	free_KRB_SAFE_BODY
	free_KX509_ERROR_CODE
	free_KerberosString
	free_KerberosTime
	free_KeyEncryptionAlgorithmIdentifier
	free_KeyIdentifier
	free_KeyTransRecipientInfo
	free_KeyUsage
	free_KrbCredInfo
	free_KrbFastArmor
	free_KrbFastArmoredRep
	free_KrbFastArmoredReq
	free_KrbFastFinished
	free_KrbFastReq
	free_KrbFastResponse
	free_Kx509Request
	free_Kx509Response
	free_LR_TYPE
	free_LastReq
	free_MESSAGE_TYPE
	free_METHOD_DATA
	free_MS_UPN_SAN
	free_MessageDigest
	free_NAME_TYPE
	free_NTLMInit
	free_NTLMInitReply
	free_NTLMReply
	free_NTLMRequest
	free_NTLMRequest2
	free_NTLMResponse
	free_Name
	free_NameConstraints
	free_OriginatorInfo
	free_OtherName
	free_PADATA_TYPE
	free_PA_DATA
	free_PA_ENC_SAM_RESPONSE_ENC
	free_PA_ENC_TS_ENC
	free_PA_FX_FAST_REPLY
	free_PA_FX_FAST_REQUEST
	free_PA_PAC_REQUEST
	free_PA_PK_AS_REP
	free_PA_PK_AS_REP_BTMM
	free_PA_PK_AS_REP_Win2k
	free_PA_PK_AS_REQ
	free_PA_PK_AS_REQ_Win2k
	free_PA_S4U2Self
	free_PA_SAM_CHALLENGE_2
	free_PA_SAM_CHALLENGE_2_BODY
	free_PA_SAM_REDIRECT
	free_PA_SAM_RESPONSE_2
	free_PA_SAM_TYPE
	free_PA_SERVER_REFERRAL_DATA
	free_PA_ServerReferralData
	free_PA_SvrReferralData
	free_PKAuthenticator
	free_PKAuthenticator_Win2k
	free_PKCS12_Attribute
	free_PKCS12_Attributes
	free_PKCS12_AuthenticatedSafe
	free_PKCS12_CertBag
	free_PKCS12_MacData
	free_PKCS12_OctetString
	free_PKCS12_PBEParams
	free_PKCS12_PFX
	free_PKCS12_SafeBag
	free_PKCS12_SafeContents
	free_PKCS8Attributes
	free_PKCS8EncryptedData
	free_PKCS8EncryptedPrivateKeyInfo
	free_PKCS8PrivateKey
	free_PKCS8PrivateKeyAlgorithmIdentifier
	free_PKCS8PrivateKeyInfo
	free_PKCS9_BMPString
	free_PKCS9_friendlyName
	free_PKIXXmppAddr
	free_PROV_SRV_LOCATION
	free_PkinitSP80056AOtherInfo
	free_PkinitSuppPubInfo
	free_Principal
	free_PrincipalName
	free_Principals
	free_ProxyCertInfo
	free_ProxyPolicy
	free_RDNSequence
	free_RSAPrivateKey
	free_RSAPublicKey
	free_Realm
	free_RecipientIdentifier
	free_RecipientInfo
	free_RecipientInfos
	free_RelativeDistinguishedName
	free_ReplyKeyPack
	free_ReplyKeyPack_Win2k
	free_SAMFlags
	free_SignatureAlgorithmIdentifier
	free_SignatureValue
	free_SignedData
	free_SignerIdentifier
	free_SignerInfo
	free_SignerInfos
	free_SubjectKeyIdentifier
	free_SubjectPublicKeyInfo
	free_TBSCRLCertList
	free_TBSCertificate
	free_TD_DH_PARAMETERS
	free_TD_INVALID_CERTIFICATES
	free_TD_TRUSTED_CERTIFIERS
	free_TGS_REP
	free_TGS_REQ
	free_TYPED_DATA
	free_Ticket
	free_TicketFlags
	free_Time
	free_TransitedEncoding
	free_TrustedCA
	free_TrustedCA_Win2k
	free_TypedData
	free_UniqueIdentifier
	free_UnprotectedAttributes
	free_ValidationParms
	free_Validity
	free_Version
	free_heim_any
	free_heim_any_set
	free_krb5int32
	free_krb5uint32
	heim_any_cmp
	initialize_asn1_error_table
	initialize_asn1_error_table_r
	int2APOptions
	int2DigestTypes
	int2DistributionPointReasonFlags
	int2FastOptions
	int2KDCOptions
	int2KeyUsage
	int2SAMFlags
	int2TicketFlags
	length_AD_AND_OR
	length_AD_IF_RELEVANT
	length_AD_INITIAL_VERIFIED_CAS
	length_AD_KDCIssued
	length_AD_LoginAlias
	length_AD_MANDATORY_FOR_KDC
	length_APOptions
	length_AP_REP
	length_AP_REQ
	length_AS_REP
	length_AS_REQ
	length_AUTHDATA_TYPE
	length_AccessDescription
	length_AlgorithmIdentifier
	length_Attribute
	length_AttributeType
	length_AttributeTypeAndValue
	length_AttributeValue
	length_AuthPack
	length_AuthPack_Win2k
	length_Authenticator
	length_AuthorityInfoAccessSyntax
	length_AuthorityKeyIdentifier
	length_AuthorizationData
	length_AuthorizationDataElement
	length_BaseDistance
	length_BasicConstraints
	length_CKSUMTYPE
	length_CMSAttributes
	length_CMSCBCParameter
	length_CMSEncryptedData
	length_CMSIdentifier
	length_CMSRC2CBCParameter
	length_CMSVersion
	length_CRLCertificateList
	length_CRLDistributionPoints
	length_CRLReason
	length_Certificate
	length_CertificateList
	length_CertificateRevocationLists
	length_CertificateSerialNumber
	length_CertificateSet
	length_Certificates
	length_ChangePasswdDataMS
	length_Checksum
	length_ContentEncryptionAlgorithmIdentifier
	length_ContentInfo
	length_ContentType
	length_DHNonce
	length_DHParameter
	length_DHPublicKey
	length_DHRepInfo
	length_DSAParams
	length_DSAPublicKey
	length_DSASigValue
	length_DigestAlgorithmIdentifier
	length_DigestAlgorithmIdentifiers
	length_DigestError
	length_DigestInfo
	length_DigestInit
	length_DigestInitReply
	length_DigestREP
	length_DigestREQ
	length_DigestRepInner
	length_DigestReqInner
	length_DigestRequest
	length_DigestResponse
	length_DigestTypes
	length_DirectoryString
	length_DistributionPoint
	length_DistributionPointName
	length_DistributionPointReasonFlags
	length_DomainParameters
	length_ECDSA_Sig_Value
	length_ECParameters
	length_ECPoint
	length_ENCTYPE
	length_ETYPE_INFO
	length_ETYPE_INFO2
	length_ETYPE_INFO2_ENTRY
	length_ETYPE_INFO_ENTRY
	length_EncAPRepPart
	length_EncASRepPart
	length_EncKDCRepPart
	length_EncKrbCredPart
	length_EncKrbPrivPart
	length_EncTGSRepPart
	length_EncTicketPart
	length_EncapsulatedContentInfo
	length_EncryptedContent
	length_EncryptedContentInfo
	length_EncryptedData
	length_EncryptedKey
	length_EncryptionKey
	length_EnvelopedData
	length_EtypeList
	length_ExtKeyUsage
	length_Extension
	length_Extensions
	length_ExternalPrincipalIdentifier
	length_ExternalPrincipalIdentifiers
	length_FastOptions
	length_GeneralName
	length_GeneralNames
	length_GeneralSubtree
	length_GeneralSubtrees
	length_HostAddress
	length_HostAddresses
	length_IssuerAndSerialNumber
	length_KDCDHKeyInfo
	length_KDCDHKeyInfo_Win2k
	length_KDCFastCookie
	length_KDCFastState
	length_KDCOptions
	length_KDC_REP
	length_KDC_REQ
	length_KDC_REQ_BODY
	length_KDFAlgorithmId
	length_KRB5PrincipalName
	length_KRB5SignedPath
	length_KRB5SignedPathData
	length_KRB_CRED
	length_KRB_ERROR
	length_KRB_PRIV
	length_KRB_SAFE
	length_KRB_SAFE_BODY
	length_KX509_ERROR_CODE
	length_KerberosString
	length_KerberosTime
	length_KeyEncryptionAlgorithmIdentifier
	length_KeyIdentifier
	length_KeyTransRecipientInfo
	length_KeyUsage
	length_KrbCredInfo
	length_KrbFastArmor
	length_KrbFastArmoredRep
	length_KrbFastArmoredReq
	length_KrbFastFinished
	length_KrbFastReq
	length_KrbFastResponse
	length_Kx509Request
	length_Kx509Response
	length_LR_TYPE
	length_LastReq
	length_MESSAGE_TYPE
	length_METHOD_DATA
	length_MS_UPN_SAN
	length_MessageDigest
	length_NAME_TYPE
	length_NTLMInit
	length_NTLMInitReply
	length_NTLMReply
	length_NTLMRequest
	length_NTLMRequest2
	length_NTLMResponse
	length_Name
	length_NameConstraints
	length_OriginatorInfo
	length_OtherName
	length_PADATA_TYPE
	length_PA_DATA
	length_PA_ENC_SAM_RESPONSE_ENC
	length_PA_ENC_TS_ENC
	length_PA_FX_FAST_REPLY
	length_PA_FX_FAST_REQUEST
	length_PA_PAC_REQUEST
	length_PA_PK_AS_REP
	length_PA_PK_AS_REP_BTMM
	length_PA_PK_AS_REP_Win2k
	length_PA_PK_AS_REQ
	length_PA_PK_AS_REQ_Win2k
	length_PA_S4U2Self
	length_PA_SAM_CHALLENGE_2
	length_PA_SAM_CHALLENGE_2_BODY
	length_PA_SAM_REDIRECT
	length_PA_SAM_RESPONSE_2
	length_PA_SAM_TYPE
	length_PA_SERVER_REFERRAL_DATA
	length_PA_ServerReferralData
	length_PA_SvrReferralData
	length_PKAuthenticator
	length_PKAuthenticator_Win2k
	length_PKCS12_Attribute
	length_PKCS12_Attributes
	length_PKCS12_AuthenticatedSafe
	length_PKCS12_CertBag
	length_PKCS12_MacData
	length_PKCS12_OctetString
	length_PKCS12_PBEParams
	length_PKCS12_PFX
	length_PKCS12_SafeBag
	length_PKCS12_SafeContents
	length_PKCS8Attributes
	length_PKCS8EncryptedData
	length_PKCS8EncryptedPrivateKeyInfo
	length_PKCS8PrivateKey
	length_PKCS8PrivateKeyAlgorithmIdentifier
	length_PKCS8PrivateKeyInfo
	length_PKCS9_BMPString
	length_PKCS9_friendlyName
	length_PKIXXmppAddr
	length_PROV_SRV_LOCATION
	length_PkinitSP80056AOtherInfo
	length_PkinitSuppPubInfo
	length_Principal
	length_PrincipalName
	length_Principals
	length_ProxyCertInfo
	length_ProxyPolicy
	length_RDNSequence
	length_RSAPrivateKey
	length_RSAPublicKey
	length_Realm
	length_RecipientIdentifier
	length_RecipientInfo
	length_RecipientInfos
	length_RelativeDistinguishedName
	length_ReplyKeyPack
	length_ReplyKeyPack_Win2k
	length_SAMFlags
	length_SignatureAlgorithmIdentifier
	length_SignatureValue
	length_SignedData
	length_SignerIdentifier
	length_SignerInfo
	length_SignerInfos
	length_SubjectKeyIdentifier
	length_SubjectPublicKeyInfo
	length_TBSCRLCertList
	length_TBSCertificate
	length_TD_DH_PARAMETERS
	length_TD_INVALID_CERTIFICATES
	length_TD_TRUSTED_CERTIFIERS
	length_TGS_REP
	length_TGS_REQ
	length_TYPED_DATA
	length_Ticket
	length_TicketFlags
	length_Time
	length_TransitedEncoding
	length_TrustedCA
	length_TrustedCA_Win2k
	length_TypedData
	length_UniqueIdentifier
	length_UnprotectedAttributes
	length_ValidationParms
	length_Validity
	length_Version
	length_heim_any
	length_heim_any_set
	length_krb5int32
	length_krb5uint32
	remove_AuthorizationData
	remove_CRLDistributionPoints
	remove_DigestAlgorithmIdentifiers
	remove_ETYPE_INFO
	remove_ETYPE_INFO2
	remove_Extensions
	remove_GeneralNames
	remove_METHOD_DATA
	remove_Principals
